mat4 rotate_x(float angle) {
    float s = sin(angle);
    float c = cos(angle);
    mat4 matrix;
    matrix[0] = vec4(1, 0, 0, 0);
    matrix[1] = vec4(0, c, s, 0);
    matrix[2] = vec4(0, -s, c, 0);
    matrix[3] = vec4(0, 0, 0, 1);
    return matrix;
}

mat4 rotate_y(float angle) {
    float s = sin(angle);
    float c = cos(angle);
    mat4 matrix;
    matrix[0] = vec4(c, 0, s, 0);
    matrix[1] = vec4(0, 1, 0, 0);
    matrix[2] = vec4(-s, 0, c, 0);
    matrix[3] = vec4(0, 0, 0, 1);
    return matrix;
}

mat4 rotate_z(float angle) {
    float s = sin(angle);
    float c = cos(angle);
    mat4 matrix;
    matrix[0] = vec4(c, s, 0, 0);
    matrix[1] = vec4(-s, c, 0, 0);
    matrix[2] = vec4(0, 0, 1, 0);
    matrix[3] = vec4(0, 0, 0, 1);
    return matrix;
}

mat4 rotate(vec3 angle) {
    return rotate_x(angle.x) * rotate_y(angle.y) * rotate_z(angle.z);
}

float opSmoothUnion(float d1, float d2, float k)
{
    float h = clamp(0.5 + 0.5*(d2-d1)/k, 0.0, 1.0);
    return mix(d2, d1, h) - k*h*(1.0-h);
}

float sdSphere(vec3 p, float s)
{
    return length(p)-s;
}

float sdRoundBox(vec3 p, vec3 b, float r)
{
    vec3 q = abs(p) - b;
    return length(max(q, 0.0)) + min(max(q.x, max(q.y, q.z)), 0.0) - r;
}

float sdCappedTorus(vec3 p, vec2 sc, float ra, float rb)
{
    p.x = abs(p.x);
    float k = (sc.y*p.x>sc.x*p.y) ? dot(p.xy, sc) : length(p.xy);
    return sqrt(dot(p, p) + ra*ra - 2.0*ra*k) - rb;
}

float sdTorus(vec3 p, vec2 t)
{
    vec2 q = vec2(length(p.xz)-t.x, p.y);
    return length(q)-t.y;
}

float sdTriPrism(vec3 p, vec2 h)
{
    vec3 q = abs(p);
    return max(q.z-h.y, max(q.x*0.866025+p.y*0.5, -p.y)-h.x*0.5);
}

float sdDeathStar(vec3 p2, float ra, float rb, float d)
{
    // sampling independent computations (only depend on shape)
    float a = (ra*ra - rb*rb + d*d)/(2.0*d);
    float b = sqrt(max(ra*ra-a*a, 0.0));

    // sampling dependant computations
    vec2 p = vec2(p2.x, length(p2.yz));
    if (p.x*b-p.y*a > d*max(b-p.y, 0.0))
    return length(p-vec2(a, b));
    else
    return max((length(p)-ra),
    -(length(p-vec2(d, 0.0))-rb));
}

float sdCapsule( vec3 p, vec3 a, vec3 b, float r )
{
    vec3 pa = p - a, ba = b - a;
    float h = clamp( dot(pa,ba)/dot(ba,ba), 0.0, 1.0 );
    return length( pa - ba*h ) - r;
}

float sdf(vec3 pos) {

    float s = 100000;

    for (int i = 0; i < 11; i++) {
        vec3 p = mat3(rotate(object_1_rotate)) * (pos - object_1_pos);
        float f = float(i) * object_1_params.y + object_1_params.z;
        float d = object_1_params.w;
        vec3 tr = vec3(0, float(i-5) * object_2_params.z, 0);
        vec3 a = vec3(object_1_params.x, sin(f) * d, cos(f) * d) + tr;

        float f2 = f + object_2_params.x;
        vec3 b = vec3(-object_1_params.x, sin(f2) * d, cos(f2) * d) + tr;

        float v = sdCapsule(p, a, b, object_2_params.y);
        s = min(s, v);
    }

    return s;

    //    float scale = torus_1_rotate.x;
    //    float size = torus_1_rotate.y;
    //    float sphere1 = sdSphere(pos - torus_1_pos *scale, size);
    //    float sphere2 = sdSphere(pos - torus_2_pos *scale, size);
    //    float sphere3 = sdSphere(pos - torus_2_rotate * scale, size);
    //    return opSmoothUnion(opSmoothUnion(sphere1, sphere2, 1), sphere3, 1);
}
